;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;                    Major mode for editing Jazz programs
;;;                                                        
;;;                   Written by Francois.Bourdoncle@ensmp.fr
;;;                                                        
;;; $Revision: 1.46 $, $Date: 1999/06/03 18:07:10 $
;;;                                                        
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(require 'cc-vars)
(require 'cc-engine)
(require 'cc-mode)
(require 'cc-cmds)
(require 'cc-align)
(require 'cc-styles)
(require 'cc-defs)
(require 'cc-menus)
(require 'cc-defs)
(require 'cc-langs)

(defvar jazz-program "jazz"
  "Jazz compiler name.")

(defvar jazz-xprogram nil
  "Jazz compiler name plus arguments")

(defvar jazz-mode-hook nil
  "Jazz mode hook.")

(defvar jazz-mode-syntax-table nil
  "Syntax table used in Jazz buffers.")

(defvar jazz-mode-map nil
  "Keymap used in Jazz buffers.")

(defvar jazz-variable-stack nil)
(defvar jazz-directory nil)
(defvar jazz-process nil)
(defvar jazz-pretty-print-flag nil)
(defvar jazz-verbose-flag nil)
(defvar jazz-optimize-flag nil)
(defvar jazz-net-list-flag t)
(defvar jazz-last-location 1)
(defconst jazz-extension ".jzz")
(defconst jazz-extension-regexp "\\.jzz$")

(defvar jazz-export-directory "."
  "Export directory for HTML version of Jazz source files.")

(defvar jazz-output-directory "."
  "Export directory for HTML version of Jazz outputs.")

(defvar jazz-file-location-regexp
  "\"\\([^\"]+\\)\":\\([0-9]+\\)\\.\\([0-9]+\\)-\\([0-9]+\\)\\.\\([0-9]+\\)")

(defvar jazz-loc-regexp
  "\\([0-9]+\\):\\([0-9]+\\)\\.\\([0-9]+\\)-\\([0-9]+\\)\\.\\([0-9]+\\)")

(defvar jazz-mode-keywords
  (list
   ;; Class declaration
   '("\\<\\(class\\|interface\\|device\\)\\>\\s-+\\<\\(\\w+\\)\\>"
     2 jazz-declaration-face)
   
   ;; Package declaration
   '("\\<package\\s-+\\([\\.a-zA-Z0-9_]+\\)"
     1 (progn
         (jazz-link-create
          (match-beginning 1) (match-end 1)
          (cons 'package (match-string-no-properties 1))
          'jazz-link-declaration-face)
         jazz-declaration-face))
   
   ;; Object import
   '("^\\s-*import\\s-+\\([*\\.a-zA-Z0-9_]+\\)"
     1 (jazz-link-create
        (match-beginning 1) (match-end 1)
        (cons 'symbol (match-string-no-properties 1))
        'jazz-link-face))
   
   ;; Package import
   '("^\\s-*require\\s-+\\([\\.a-zA-Z0-9_]+\\)"
     1 (jazz-link-create
        (match-beginning 1) (match-end 1)
        (cons 'package (match-string-no-properties 1))
        'jazz-link-face))
   
   ;; Method declaration
   '("^\\s-*\\(\\<\\(static\\|public\\|fun\\|final\\|native\\)\\>\\s-+\\)*\\(\\<\\w+\\>\\.\\)?\\(\\<\\w+\\>\\)\\s-*\\(<\\(\\w\\|\\s-\\|[,:]\\)*>\\)?\\s-*\\(@\\s-*\\w+\\(<\\(\\w\\|\\s-\\|[,:]\\)*>\\)?\\)?\\s-*([^#!@%&\\*\\+-/<>=\\^|~]"
     4 jazz-declaration-face)
   
   ;; Bogus method declaration
   '("\\<\\(native\\)\\s-*("
     1 nil t)
   
   ;; Bogus method declaration
   '("[-?:,=*+/)(]\\s-*
+\\s-*\\(\\<[a-z][A-Z_a-z0-9']*\\>\\.\\)?\\<\\(\\w+\\)\\>\\s-*("
     2 nil t)
   
   ;; Method declaration after comments
   '("^\\s-*//.*
+\\s-*\\<\\(\\w+\\)\\>\\s-*("
     1 jazz-declaration-face t)
   
   ;; Field declaration
   '("^\\s-*\\(\\<\\(static\\|input\\|output\\|public\\|dynamic\\)\\>\\s-+\\)*\\<\\(\\w+\\)\\>\\s-*:.*[=;]"
     3 jazz-declaration-face)
   
   ;; Variable declaration
   '("^\\s-*\\<var\\>\\s-+\\<\\(\\w+\\)\\>\\s-*\\(:\\s-*[^=;]*\\)?\\s-*[;=]"
     1 jazz-declaration-face)
   
   ;; Bogus field declaration
   '("[-?:,=*+/)(]\\s-*
+\\s-*\\<\\(\\w+\\)\\>\\s-*:"
     1 nil t)
   
   ;; Field declaration after comments
   '("^\\s-*//.*
+\\s-*\\<\\(\\w+\\)\\>\\s-*:"
     1 jazz-declaration-face t)
   
   ;; Constants
   '("\\<\\(true\\|false\\|this\\|[A-Z][A-Z_]+\\)\\>"
     1 jazz-constant-face)
   
   ;; Classes
   '("[^\\.]\\<\\(_?[A-Z][a-zA-Z][a-zA-Z_'0-9]*\\|float\\|per\\|rat\\|int\\|boolean\\)\\>"
     1 jazz-type-face)
   
   ;; Classes
   '("^\\<\\(_?[A-Z][a-zA-Z][a-zA-Z_'0-9]*\\|float\\|per\\|rat\\|int\\|boolean\\)\\>"
     1 jazz-type-face)
   
   ;; Field definition
   '("^\\s-*_?[A-Z][a-zA-Z_][a-zA-Z_'0-9]*\\.\\(\\w+\\)\\s-*="
     1 jazz-declaration-face)
   
   ;; Keywords
   '("\\<\\(pragma\\|format\\|fun\\|print\\|error\\|static\\|final\\|extends\\|implements\\|abstract\\|abstracts\\|public\\|var\\|dynamic\\|class\\|interface\\|device\\|new\\|device\\|as\\|export\\|instanceof\\|for\\|if\\|else\\|input\\|output\\|div\\|native\\|extern\\|reset\\|enable\\|import\\|require\\|package\\|alike\\)\\>\\|@"
     0 jazz-keyword-face)
   
   ;; Constraints
   '("\\({[^{}]*}\\)\\s-*[;=]"
     1 jazz-constraint-face t)
   
   ;; For/if/assert
   '("^\\s-*\\(dynamic\\|assert\\|for\\|reset\\|enable\\|if\\)\\s-*("
     1 jazz-keyword-face t))
  "Jazz mode keywords.")

(defvar cc-imenu-jazz-generic-expression
  (`
   ((nil
     (,
      (concat
       "^\\([ \t]\\)*"
       "\\([A-Za-z0-9_-]+[ \t]+\\)?"    ; type specs; there can be
       "\\([A-Za-z0-9_-]+[ \t]+\\)?"    ; more than 3 tokens, right?
       "\\([A-Za-z0-9_-]+[ \t]*[[]?[]]?\\)"
       "\\([ \t]\\)"
       "\\([A-Za-z0-9_-]+\\)"           ; the string we want to get
       "\\([ \t]*\\)+("
       "\\([a-zA-Z,_1-9\n \t]*[[]?[]]?\\)*" ; arguments
       ")[ \t]*"
       "[^;(]"
       "[,a-zA-Z_1-9\n \t]*{"               
       )) 6)))
  "Imenu generic expression for Jazz mode.  See `imenu-generic-expression'.")

;; Face used to select statements/expressions
(make-face 'jazz-selection-face)
(set-face-background 'jazz-selection-face "#cccccc")
(set-face-foreground 'jazz-selection-face "black")

;; Face used to emphasize keywords
(make-face 'jazz-keyword-face)
(set-face-foreground 'jazz-keyword-face "#0000ee")
(defconst jazz-keyword-face 'jazz-keyword-face)

;; Face used to emphasize constants
(make-face 'jazz-constant-face)
(set-face-foreground 'jazz-constant-face "#8080ff")
(defconst jazz-constant-face 'jazz-constant-face)

;; Face used to emphasize string constants
(make-face 'jazz-string-face)
(set-face-foreground 'jazz-string-face "#a020f0")
(defconst jazz-string-face 'jazz-string-face)

;; Face used to emphasize declarations
(make-face 'jazz-declaration-face)
(set-face-foreground 'jazz-declaration-face "#cd0000")
(defconst jazz-declaration-face 'jazz-declaration-face)

;; Face used to emphasize types
(make-face 'jazz-type-face)
(set-face-foreground 'jazz-type-face "#b7860b")
(defconst jazz-type-face 'jazz-type-face)

;; Face used to emphasize constraints
(make-face 'jazz-constraint-face)
(set-face-foreground 'jazz-constraint-face "#e9967a")
(defconst jazz-constraint-face 'jazz-constraint-face)

;; Face used to emphasize comments
(make-face 'jazz-comment-face)
(set-face-foreground 'jazz-comment-face "#008b00")
(defconst jazz-comment-face 'jazz-comment-face)

;; Face used to emphasize links
(make-face 'jazz-link-face)
(set-face-foreground 'jazz-link-face "#000000")
(set-face-underline-p 'jazz-link-face t)
(defconst jazz-link-face 'jazz-link-face)

;; Face used to emphasize declaration links
(make-face 'jazz-link-declaration-face)
(set-face-foreground 'jazz-link-declaration-face "#cd0000")
(set-face-underline-p 'jazz-link-declaration-face t)
(defconst jazz-link-declaration-face 'jazz-link-declaration-face)

;; System-dependant path separator
(defconst jazz-path-separator
  (if (eq window-system 'w32) ";" ":"))

;; Syntax table initialization
(if jazz-mode-syntax-table
    ()
  (setq jazz-mode-syntax-table (make-syntax-table))
  (c-populate-syntax-table jazz-mode-syntax-table)
  (modify-syntax-entry ?_ "w" jazz-mode-syntax-table)
  (modify-syntax-entry ?\' "w" jazz-mode-syntax-table))

;; Local keymap initialization

(if jazz-mode-map
    ()
  (setq jazz-mode-map (make-sparse-keymap))
  (define-key jazz-mode-map [menu-bar] (make-sparse-keymap))
  (define-key jazz-mode-map [menu-bar c]
    (cons "Jazz" (make-sparse-keymap "Jazz")))
  (define-key jazz-mode-map [menu-bar c jazz-toggle-net-list]
    '("Toggle net-list mode" . jazz-toggle-net-list))
  (define-key jazz-mode-map [menu-bar c jazz-toggle-pretty-print]
    '("Toggle pretty-print mode" . jazz-toggle-pretty-print))
  (define-key jazz-mode-map [menu-bar c jazz-toggle-verbose]
    '("Toggle verbose mode" . jazz-toggle-verbose))
  (define-key jazz-mode-map [menu-bar c jazz-toggle-optimize]
    '("Toggle optimize mode" . jazz-toggle-optimize))
  (define-key jazz-mode-map [menu-bar c jazz-html-listing]
    '("Generate HTML listing" . jazz-html-listing))
  (define-key jazz-mode-map [menu-bar c jazz-latex-listing]
    '("Generate LaTeX listing" . jazz-latex-listing))
  (define-key jazz-mode-map [menu-bar c jazz-display-package]
    '("Display package" . jazz-display-package))
  (define-key jazz-mode-map [menu-bar c jazz-uncomment-region-or-line]
    '("Uncomment region/line" . jazz-uncomment-region-or-line))
  (define-key jazz-mode-map [menu-bar c jazz-comment-region-or-line]
    '("Comment region/line" . jazz-comment-region-or-line))
  (define-key jazz-mode-map [menu-bar c jazz-next-error]
    '("Next error" . jazz-next-error))
  (define-key jazz-mode-map [menu-bar c jazz-compile-buffer]
    '("Compile buffer" . jazz-compile-buffer)))

;; Jazz syntactic constructs
(defconst c-Jazz-conditional-key
  "\\b\\(for\\|if\\|do\\|else\\|while\\|switch\\)\\b[^_]")

(defconst c-Jazz-comment-start-regexp "/\\(/\\|[*][*]?\\)")

(defconst c-Jazz-class-key
  (concat
   "\\(" c-protection-key "\\s +\\)?"
   "\\(interface\\|class\\|device\\)\\s +"
   c-symbol-key                         ;name of the class
   "\\(\\s *extends\\s *" c-symbol-key "\\)?" ;maybe followed by superclass 
   "\\(\\s *implements *[^{]+{\\)?"     ;maybe the adopted protocols list
   ))

(defconst c-Jazz-method-key
  (concat
   "^\\s *[+-]\\s *"
   "\\(([^)]*)\\)?"			; return type
   ;; \\s- in jazz syntax table does not include \n
   ;; since it is considered the end of //-comments.
   "[ \t\n]*" c-symbol-key))

(defconst c-Jazz-access-key nil)

(defun jazz-mode ()
  "
Mode for editing/compiling Jazz programs and browsing net lists.

* PROGRAMS CAN BE EDITED WITH THE FOLLOWING COMMANDS:

  Key                     Command
  ---                     -------
  
  Tab                     indent-for-tab-command
  Help                    jazz-next-error
  C-c C-n                 jazz-next-error
  Do                      jazz-compile-buffer
  C-c C-b                 jazz-compile-buffer
  C-c C-h                 jazz-html-listing
  C-c C-l                 jazz-latex-listing
  C-c C-p                 jazz-display-package
  C-c C-c                 jazz-comment-region-or-line
  C-c C-u                 jazz-uncomment-region-or-line
  C-c n                   jazz-toggle-net-list
  C-c p                   jazz-toggle-pretty-print
  C-c v                   jazz-toggle-verbose
  C-c o                   jazz-toggle-optimize

* NET LISTS CAN BE INSPECTED WITH THE FOLLOWING COMMANDS:

  Key                     Command
  ---                     -------
  
  Tab                     jazz-find-variable
  S-Tab                   jazz-previous-location
  Space                   jazz-locate

* THE MODE IS CONTROLED BY THE FOLLOWING VARIABLES:

  jazz-program            Name of jazz compiler (default = \"jazz\")
  jazz-mode-hook          User mode hook called after initialization

* COLORS AND FONTS CAN BE CUSTOMIZED WITH THE FOLLOWING FACES:

  Faces
  -----

  jazz-selection-face     Face used to select source code
  jazz-keyword-face       Face used to emphasize keywords
  jazz-constant-face      Face used to emphasize constants
  jazz-declaration-face   Face used to emphasize declarations
  jazz-string-face        Face used to emphasize string constants
  jazz-type-face          Face used to emphasize types
  jazz-constraint-face    Face used to emphasize constraints
  jazz-comment-face       Face used to emphasize comments
  jazz-link-face          Face used to highlight links
  jazz-link-declaration-face Face used to highlight declaration links"

  (interactive)
  (c-initialize-cc-mode)
  (kill-all-local-variables)
  (set-syntax-table jazz-mode-syntax-table)
  (setq major-mode 'jazz-mode
 	mode-name "Jazz"
 	local-abbrev-table java-mode-abbrev-table)
  (use-local-map jazz-mode-map)
  (c-common-init)
  (setq comment-start "// "
 	comment-end   ""
 	c-conditional-key c-Jazz-conditional-key
 	c-comment-start-regexp c-Jazz-comment-start-regexp
  	c-class-key c-Jazz-class-key
	c-method-key c-Jazz-method-key
 	c-baseclass-key nil
	c-recognize-knr-p nil
 	c-access-key c-Jazz-access-key
        c-label-key "###"
	imenu-generic-expression cc-imenu-jazz-generic-expression)
  (c-add-style
   "jazz"
   '((c-basic-offset . 2)
     (c-comment-only-line-offset 0 . 0)
     (c-hanging-comment-starter-p)
     (c-offsets-alist
      (topmost-intro-cont . +)
      (statement-block-intro . +)
      (knr-argdecl-intro . 5)
      (substatement-open . +)
      (label . +)
      (statement-case-open . +)
      (statement-cont . jazz-lineup-runin-statements)
      (arglist-intro . c-lineup-arglist-intro-after-paren)
      (arglist-close . c-lineup-arglist)
      (arglist-cont . c-lineup-arglist-intro-after-paren)
      (access-label . +)
      (inher-cont . c-lineup-java-inher)
      (func-decl-cont . c-lineup-java-throws))) t)
  (local-set-key "\C-c\C-b" 'jazz-compile-buffer)
  (local-set-key "\C-c\C-c" 'jazz-comment-region-or-line)
  (local-set-key "\C-c\C-u" 'jazz-uncomment-region-or-line)
  (local-set-key "\C-c\C-h" 'jazz-html-listing)
  (local-set-key "\C-c\C-l" 'jazz-latex-listing)
  (local-set-key "\C-c\C-p" 'jazz-display-package)
  (local-set-key "\C-c\C-n" 'jazz-next-error)
  (local-set-key "\C-c\C-e" 'jazz-export-file)
  (local-set-key "\C-cp" 'jazz-toggle-pretty-print)
  (local-set-key "\C-co" 'jazz-toggle-optimize)
  (local-set-key "\C-cn" 'jazz-toggle-net-list)
  (local-set-key "\C-cv" 'jazz-toggle-verbose)
  (local-set-key "\C-l" 'jazz-fontify-buffer)
  (local-set-key "{" 'c-electric-brace)
  (local-set-key "}" 'c-electric-brace)
  (local-set-key [menu] 'jazz-compile-buffer)
  (local-set-key [mouse-1] 'jazz-link-activate)
  (local-set-key [down-mouse-3] 'jazz-files-menu)
  (local-set-key [help] 'jazz-next-error)
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults
        '(jazz-mode-keywords))
  (make-local-variable 'font-lock-comment-face)
  (setq font-lock-comment-face 'jazz-comment-face)
  (make-local-variable 'font-lock-string-face)
  (setq font-lock-string-face 'jazz-string-face)
  (turn-on-font-lock)
  (setq buffer-modified-p nil)
  (let ((name (buffer-file-name)))
    (when name
      (save-excursion
        (beginning-of-buffer)
        (cond
         ((re-search-forward
           "^package\\s-+\\([\\.a-zA-Z0-9_]*\\)\\s-*[{;]" (point-max) t)
          (jazz-rename-buffer (concat (match-string-no-properties 1) "."
                                      (file-name-sans-extension
                                       (file-name-nondirectory
                                        (buffer-file-name))))))
         ((jazz-file-p name)
          (jazz-rename-buffer (file-name-nondirectory
                               (substring name 0 (- (length name) 4)))))))))
  (run-hooks 'jazz-mode-hook)
  (c-update-modeline))

(defun jazz-lineup-runin-statements (langelem)
  ;; line up statements in coding standards which place the first
  ;; statement on the same line as the block opening brace.
  (let ((char  (cdr langelem)))
    (save-excursion
      (goto-char (cdr langelem))
      (if (looking-at ".*=$") 2 0))))

(defun jazz-compile-buffer ()
  "Compile a jazz buffer."
  (interactive)
  (let (cmd)
    (setq jazz-last-location 1)
    (setq jazz-variable-stack nil)
    (if jazz-process
        (delete-process jazz-process))
    (if (get-buffer "*Jazz*")
        (kill-buffer "*Jazz*"))
    (save-buffer)

    ;; Build process command
    (setq cmd
          (append
           (if jazz-xprogram jazz-xprogram (cons jazz-program nil))
           (if jazz-pretty-print-flag '("-p" "-t"))
           (if jazz-net-list-flag '("-o" "-"))
           (if jazz-verbose-flag '("-v"))
           (if jazz-optimize-flag '("-O"))
           (cons (file-name-nondirectory (buffer-file-name))
                 nil)))
    
    ;; Save the current directory
    (setq jazz-directory (file-name-directory (buffer-file-name)))

    ;; Start process
    (setq jazz-process (jazz-start-process jazz-directory cmd))

    ;; User feedback
    (setq jazz-compiling t)
    
    ;; Bindings
    (let ((name (jazz-buffer-name (current-buffer))))
      (save-excursion
        (set-buffer "*Jazz*")
        (set (make-variable-buffer-local 'jazz-output-name) name)
        (delete-region (point-min) (point-max))
        (local-set-key " " 'jazz-find-variable)
        (local-set-key [S-tab] 'jazz-previous-location)
        (local-set-key "\C-c\C-e" 'jazz-export-output)
        (local-set-key " " 'jazz-locate)))

    ;; Display the compilation buffer
    (display-buffer "*Jazz*")))

(defun jazz-next-error ()
  "Find next Jazz error."
  (interactive)
  (let ((buffer (get-buffer "*Jazz*"))
        start file l1 c1 l2 c3)
    (if (not buffer)
        (message "Error: buffer not compiled." (beep))
      (if jazz-last-location
          (save-excursion
            (set-buffer buffer)
            (setq mark-active nil)
            (goto-char jazz-last-location)
            (if (setq jazz-last-location
                      (jazz-search-forward jazz-file-location-regexp
                                           (point-max)))
                (progn
                  (setq start (match-beginning 1))
                  (setq file (match-string-no-properties 1))
                  (setq l1 (jazz-match-to-int 2))
                  (setq c1 (jazz-match-to-int 3))
                  (setq l2 (jazz-match-to-int 4))
                  (setq c2 (jazz-match-to-int 5)))
              (jazz-no-more-errors)))
        (jazz-no-more-errors))
      (if (and file l1 c1 l2 c2)
          (progn
            (jazz-select-file file nil l1 c1 l2 c2)
            (set-window-start (display-buffer buffer)
                              (save-excursion
                                (set-buffer buffer)
                                (goto-char start)
                                (beginning-of-line)
                                (point))))
        (jazz-clear)))))

(cond
 ;; Hilighting for XEmacs
 ((string-match "XEmacs" emacs-version) 
  (defun match-string-no-properties (n) 
    (match-string n)) 
  (defvar jazz-selection-extent (make-extent 1 1)) 
  
  (defun jazz-select (window start end)  
    (set-extent-endpoints  
     jazz-selection-extent  
     start  
     end  
     (window-buffer window)) 
    (if (not pre-command-hook) 
        (set-extent-face jazz-selection-extent 'jazz-selection-face)) 
    (set-window-point window start) 
    (setq pre-command-hook 'jazz-clear)) 
  
  (defun jazz-clear () 
    (set-extent-face jazz-selection-extent 'default) 
    (setq pre-command-hook nil)))

 ;; Hilighting for GNU Emacs
 (t 
  (defvar jazz-overlay (make-overlay 1 1)) 
  (defun jazz-clear () 
    (overlay-put jazz-overlay 'face 'default) 
    (setq pre-command-hook nil)) 
  
  
  (defun jazz-select (window start end) 
    (move-overlay jazz-overlay start end (window-buffer window)) 
    (if (not pre-command-hook) 
        (overlay-put jazz-overlay 'face 'jazz-selection-face)) 
    (set-window-point window start) 
    (setq pre-command-hook 'jazz-clear))))

(defun jazz-newline ()
  "Indent the line, then insert a newline."
  (interactive)
  (indent-for-tab-command)
  (newline-and-indent))

(defun jazz-toggle-pretty-print ()
  "Disable/enable pretty-print of Jazz programs."
  (interactive)
  (setq jazz-pretty-print-flag (not jazz-pretty-print-flag))
  (message "Pretty print %s." (if jazz-pretty-print-flag "on" "off")))

(defun jazz-toggle-verbose ()
  "Disable/enable the verbose mode."
  (interactive)
  (setq jazz-verbose-flag (not jazz-verbose-flag))
  (message "Verbose mode %s." (if jazz-verbose-flag "on" "off")))

(defun jazz-toggle-optimize ()
  "Disable/enable the optimization of the netlist (alias reduction)."
  (interactive)
  (setq jazz-optimize-flag (not jazz-optimize-flag))
  (message "Optimization %s." (if jazz-optimize-flag "on" "off")))

(defun jazz-toggle-net-list ()
  "Disable/enable the insertion of the net list in the output buffer."
  (interactive)
  (setq jazz-net-list-flag (not jazz-net-list-flag))
  (if jazz-net-list-flag 
      (message "Net-list printed to standard output.")
    (message "Net-list saved to disk.")))

;; Return the package of a object (nil for global objects)
(defun jazz-object-package (object)
  (let ((package
         (mapconcat 'identity
                    (reverse
                     (cdr
                      (reverse
                       (split-string object "\\.")))) ".")))
    (if (string-equal package "") nil package)))

;; Return the relative path of an object
(defun jazz-object-directory (object)
  (let ((package
         (mapconcat 'identity
                    (reverse
                     (cdr
                      (reverse
                       (split-string object "\\.")))) "/")))
    (if (string-equal package "") "." package)))

;; Return the object without its package qualification
(defun jazz-object-name (object)
  (car (last (split-string object "\\."))))

;; Find the object "oname" of the kind "kind" in package "opackage" (directory
;; "odir" is the relative path corresponding to package "opackage")
(defun jazz-find-object-in-dir (dir odir kind opackage oname)
  (let ((object (concat dir "/" odir "/" oname)))
    (cond
     ((and (eq kind 'package)
           (jazz-package-p object))
      (list 'package opackage oname object))
     ((and (eq kind 'file)
           (jazz-file-p (concat object jazz-extension)))
      (list 'file opackage oname (concat object jazz-extension)))
     ((eq kind 'symbol)
      (or
       ;; Try to locate a file with the name of the symbol
       (jazz-find-object-in-dir dir odir 'file opackage oname)
       
       ;; Try to locate a file with the name of the symbol in lowercase
       (jazz-find-object-in-dir dir odir 'file opackage (downcase oname))
       
       ;; Otherwise display the package of the symbol
       (and opackage
            (jazz-find-object-in-dir dir
                                     (file-name-directory odir)
                                     'package
                                     (jazz-object-package opackage)
                                     (jazz-object-name opackage)))))
     (t nil))))

;; Find the object of the given kind (file/package/symbol)
(defun jazz-find-object (kind object)
  (let* ((opackage (jazz-object-package object))
         (odir (jazz-object-directory object))
         (oname (jazz-object-name object))
         (path (cons "." (split-string (or (getenv "JAZZPATH") ".")
                                       jazz-path-separator))))
    (catch 'found
      (while path
        (let ((job (jazz-find-object-in-dir
                    (car path) odir kind opackage oname)))
          (if job (throw 'found job))
          (setq path (cdr path)))))))
  
;; Create a link
(defun jazz-link-create (beg end link face)
  (when (jazz-find-object (car link) (cdr link))
    (put-text-property beg end 'mouse-face face)
    (put-text-property beg end 'front-nonsticky t)
    (put-text-property beg end 'rear-nonsticky t)
    (put-text-property beg end 'jazz-link link)
    nil))

;; Activate a link
(defun jazz-link-activate ()
  (interactive)
  (deactivate-mark)
  (let* ((link (get-text-property (point) 'jazz-link)))
    (when link (jazz-display-object (car link) (cdr link) nil))))

;; Display a given package
(defun jazz-display-package (package)
  "Display a Jazz package."
  (interactive "sPackage: ")
  (jazz-display-object 'package package nil))

;; Returns true if "dir" is a Jazz package
(defun jazz-package-p (dir)
  (let* ((name (file-name-nondirectory dir))
         (dir (concat dir "/.jazz")))
    (and (not (string-equal name "."))
         (not (string-equal name ".."))
         (file-directory-p dir))))

;; Returns true if "file" is a jazz source file
(defun jazz-file-p (file)
  (and file
       (file-readable-p file)
       (string-match jazz-extension-regexp file)))

;; Make the specified object of the given kind (file/package/symbol)
;; visible in some window
(defun jazz-display-object (kind object export)
  (let* ((job (jazz-find-object kind object))
         (kind (and job (nth 0 job)))
         (opackage (and job (nth 1 job)))
         (oname (and job (nth 2 job)))
         (opath (and job (nth 3 job))))
    (cond
     ;; Package
     ((eq kind 'package)
      (let* ((allfiles (directory-files opath t nil nil))
             (package (if opackage (concat opackage "." oname) oname))
             (buffer "*Jazz package*")
             beg end files state)
        ;; Create a temporary buffer
        (with-output-to-temp-buffer buffer
          (save-excursion
            (set-buffer buffer)
            (toggle-read-only -1)
            (setq beg (point))
            (princ "package ")
            (put-text-property beg (1- (point)) 'face jazz-keyword-face)
            (setq beg (point))
            (if (not package)
                (princ "root")
              (let ((p (jazz-object-package package))
                    (n (jazz-object-name package)))
                (if (not p)
                    (princ package)
                  (princ p)
                  (jazz-link-create beg (point)
                                    (cons 'package p)
                                    'jazz-link-declaration-face)
                  (princ ".")
                  (princ n))))
            (setq end (point))
            (princ " {\n")
            (put-text-property beg end 'face jazz-declaration-face)
            (setq state 0)
            (while (< state 2)
              (setq files allfiles)
              (while files
                (let* ((file (and (file-readable-p (car files)) (car files)))
                       (name (and file (file-name-nondirectory file)))
                       (ispackage (and file (jazz-package-p file)))
                       (isfile (and file (jazz-file-p file))))
                  (when (or (and ispackage (= state 0))
                            (and isfile (= state 1)))
                    (princ "  ")
                    (setq beg (point))
                    (princ (if isfile "file" "package"))
                    (princ " ")
                    (put-text-property beg (1- (point))
                                       'face jazz-keyword-face)
                    (setq beg (point))
                    (if ispackage
                        (when package
                          (princ package)
                          (princ ".")))
                    (let ((nm
                           (if ispackage name
                             (substring name 0 (- (length name) 4)))))
                      (princ nm)
                      (jazz-link-create
                       beg (point)
                       (cons (if ispackage 'package 'file)
                             (concat (if package
                                         (concat package ".") "") nm))
                       'jazz-link-face))
                    (princ ";\n"))
                  (setq files (cdr files))))
              (setq state (1+ state)))
            (princ "}")))
        
        ;; Bindings for the temporary buffer
        (let ((name (concat
                     "package."
                     (if (not package) "root"
                       (let ((p (jazz-object-package package))
                             (n (jazz-object-name package)))
                         (if (not p) package (concat p "." n)))))))
          (save-excursion
            (set-buffer buffer)
            (if (get-buffer name)
                (kill-buffer name))
            (jazz-rename-buffer name)
            (local-set-key [mouse-1] 'jazz-link-activate)
            (local-set-key "\C-c\C-h" 'jazz-html-listing)
            (local-set-key "\C-c\C-l" 'jazz-latex-listing)
            (local-set-key "\C-c\C-e" 'jazz-export-package)
            (when export (jazz-export ""))))))

     ;; Source file
     ((eq kind 'file)
      (let ((window (display-buffer (find-file-noselect opath t nil))))
        (save-excursion
          (set-buffer (window-buffer window))
          (beginning-of-buffer)
          (let ((case-fold-search t))
            (when (re-search-forward (concat "^\\s-*\\<public\\>.*\\(\\<"
                                             oname "\\>\\)")
                                     (point-max) t nil)
              (jazz-select window (match-beginning 1)
                           (match-end 1)))))))
     (t
      (message "Object \"%s\" not found." object)))))

(defun jazz-rename-buffer (name)
  (if (get-buffer name)
      (let ((n 2) tmp)
        (while (get-buffer (setq tmp (concat name "<" n ">")))
          (setq n (+ n 1)))
        (setq name tmp)))
  (rename-buffer name))

(defun jazz-buffer-name (buffer)
  (let ((name (buffer-name buffer)))
    (when name
      (if (string-match "<[0-9]+>$" name)
          (substring name 0 (match-beginning 0))
        name))))

(defun jazz-files-menu (event)
  "Pop up a menu of Jazz modules for selection with the mouse."
  (interactive "e")
  (let ((buffer-list (buffer-list))
        entry-list buffers menu)
    ;; Opened modules
    (while buffer-list
      (let* ((buffer (car buffer-list))
             (file (buffer-file-name buffer)))
        (cond
         ((jazz-file-p file)
          (setq entry-list (cons (cons (buffer-name buffer) buffer)
                                 entry-list)))
         ((string-match "^package\\." (buffer-name buffer))
          (setq entry-list (cons (cons (buffer-name buffer) buffer)
                                 entry-list))))
        (setq buffer-list (cdr buffer-list))))
    ;; Sort the menu
    (setq entry-list
          (sort entry-list
                '(lambda (x y)
                   (string-lessp (car x) (car y)))))
    ;; Display the menu
    (if (> (length entry-list) 0)
        (let* ((menu (list "Jazz files" (cons "" entry-list)))
               (buf (x-popup-menu event menu))
               (window (posn-window (event-start event))))
          (when buf
            (or (framep window) (select-window window))
            (switch-to-buffer buf))))))

(defun jazz-previous-location ()
  "Jump back to the initial location after a jazz-find-variable."
  (interactive)
  (if (not jazz-variable-stack)
      (message "No previous variable." (beep))
    (goto-char (car jazz-variable-stack))
    (setq jazz-variable-stack (cdr jazz-variable-stack))
    (jazz-locate)))
  
(defun jazz-find-variable ()
  "Locate a variable definition."
  (interactive)
  (let (point eol bol var found input)
    (setq point (point))
    (save-excursion
      (beginning-of-line)
      (setq bol (point))
      (goto-char point)
      (if (and (jazz-search-backward "[^0-9]" bol)
               (looking-at ".\\([0-9]+\\)"))
          (progn
            (setq var (match-string-no-properties 1))
            (beginning-of-buffer)
            (if (jazz-search-forward (concat "^ " var ":.* = ") (point-max))
                (setq found (match-beginning 0))
              (progn
                (beginning-of-buffer)
                (setq input (concat "^inputs {[^}]* " var " "))
                (if (jazz-search-forward input (point-max))
                    (progn
                      (message "Input variable." (beep))
                      (setq input t))))))))
    (if (not found)
        (if (not input)
            (message "Definition not found." (beep)))
      (progn
        (setq jazz-variable-stack (cons point jazz-variable-stack))
        (goto-char found)
        (jazz-locate)))))

(defun jazz-comment-region-or-line ()
   "Comment the region, or the line if the mark is not active."
   (interactive)
   (if mark-active
       (jazz-comment-region (min (point) (mark))
			  (max (point) (mark)))
     (jazz-comment-region
      (save-excursion (beginning-of-line) (point))
      (save-excursion (end-of-line) (point)))))

(defun jazz-uncomment-region-or-line ()
   "Uncomment the region, or the line if the mark is not active."
   (interactive)
   (if mark-active
       (jazz-uncomment-region (min (point) (mark))
			    (max (point) (mark)))
     (jazz-uncomment-region
      (save-excursion (beginning-of-line) (point))
      (save-excursion (end-of-line) (point)))))

(defun jazz-comment-region (start end)
  (save-excursion
    (goto-char end)
    (setq end (point-marker))
    (save-restriction
      (narrow-to-region
       (progn (goto-char start) (beginning-of-line) (point))
       (progn (goto-char end) (or (bolp) (forward-line 1)) (point)))
      (goto-char (point-min))
      (while (not (eobp))
	(insert "//")
	(forward-line 1))))
  (indent-region start end nil))

(defun jazz-uncomment-region (start end)
  (save-excursion
    (goto-char end)
    (setq end (point-marker))
    (save-restriction
      (narrow-to-region
       (progn (goto-char start) (beginning-of-line) (point))
       (progn (goto-char end) (forward-line 1) (point)))
      (goto-char (point-min))
      (let ((comment-regexp "\\s-*//"))
	(while (not (eobp))
	  (if (looking-at comment-regexp)
	      (delete-region (match-beginning 0) (match-end 0)))
	  (forward-line 1)))))
  (indent-region start end nil))

(defun jazz-fontify-buffer ()
  "Fontifies the current Jazz buffer."
  (interactive)
  (let ((modified (buffer-modified-p)))
    (remove-text-properties (point-min) (point-max) '(mouse-face))
    (font-lock-fontify-buffer)
    (set-buffer-modified-p modified)))

(defun jazz-html-link (beg end)
  (let ((link (get-text-property beg 'jazz-link)))
    (when link
      (let* ((job (jazz-find-object (car link) (cdr link)))
             (kind (and job (nth 0 job)))
             (opackage (and job (nth 1 job)))
             (oname (and job (nth 2 job)))
             (opath (and job (nth 3 job))))
        (cond
         ((eq kind 'package)
          (concat "package."
                  (if opackage (concat opackage ".") "")
                  oname ".html"))
         ((eq kind 'file)
          (concat (if opackage (concat opackage ".") "")
                  oname jazz-extension ".html"))
         (t "undefined.html"))))))

(defun jazz-export (ext)
  (let* ((html "*jazz-html-listing*")
         (file (concat jazz-export-directory
                       "/" (concat (jazz-buffer-name (current-buffer))
                                   ext ".html"))))
    (save-excursion
      (if (get-buffer html)
          (kill-buffer html))
      (jazz-html-listing)
      (set-buffer html)
      (write-file file nil)
      (set-file-modes file 420)
      (kill-buffer nil))))

(defun jazz-export-file ()
  (interactive)
  (jazz-export jazz-extension))
  
(defun jazz-export-package ()
  (interactive)
  (jazz-export ""))

(defun jazz-export-all-packages-in-dir (dir prefix)
  (let ((files (directory-files dir t nil nil)))
    (while files
      (let ((file (car files)))
        ;; Source file
        (when (jazz-file-p file)
          (save-excursion
            (set-buffer (find-file-noselect file))
            (jazz-export jazz-extension)))
        
        ;; Sub-package
        (when (jazz-package-p file)
          (let* ((name (file-name-nondirectory file))
                 (package (if prefix (concat prefix "." name) name)))
            (jazz-display-object 'package package t)
            (jazz-export-all-packages-in-dir (car files) package))))
      (setq files (cdr files)))))
  
(defun jazz-export-all-packages ()
  (interactive)
  (let ((path (cons "." (split-string (or (getenv "JAZZPATH") ".")
                                      jazz-path-separator)))
        packages)
    (while path
      (jazz-export-all-packages-in-dir (car path) nil)
      (setq path (cdr path)))))
                                 
(defun jazz-export-output ()
  (interactive)
  (let* ((output (buffer-string))
         (name jazz-output-name)
         (html "*jazz-html-listing*")
         (file (concat jazz-output-directory
                       "/" (concat jazz-output-name ".out.html"))))
    (save-excursion
      (set-buffer (get-buffer-create html))
      (delete-region (point-min) (point-max))
      (insert "<HTML>\n<HEAD><TITLE>Output of &quot;")
      (insert name)
      (insert "&quot;</TITLE></HEAD>\n<BODY BGCOLOR=\"#ffffff\">\n<PRE>\n")
      (insert output)
      (insert "</PRE>\n</BODY>\n</HTML>")
      (write-file file nil)
      (set-file-modes file 420)
      (kill-buffer nil))))

(defun jazz-html-listing ()
  "Formats current buffer in HTML"
  (interactive)
  (when (buffer-file-name) (jazz-fontify-buffer))
  (let ((temp-buffer-show-function '(lambda (buffer))))
    (with-output-to-temp-buffer "*jazz-html-listing*"
      (princ "<HTML><BODY BGCOLOR=\"#ffffff\"><PRE>")
      (let ((close nil)
            (beg 1)
            (end 0)
            face spec)
        (while (< end (point-max))
          (setq end (or (next-property-change beg) (point-max)))
          
          (let* ((face (get-text-property beg 'face))
                 (link (jazz-html-link beg end))
                 (entry (or (and face
                                 (assoc face jazz-html-faces-alist))
                            (and link
                                 (cons
                                  t
                                  (cons
                                   (concat
                                    "<A HREF=\"" link
                                    "\"><FONT COLOR=\"#000000\">")
                                   "</A></FONT>")))))
                 (spec (and entry (cdr entry))))
            (jazz-html-print-segment close (and spec (car spec)) beg end)
            (setq close (and spec (cdr spec)))
            (setq beg end)))
        (if close (princ close)))
      (princ "</PRE></BODY></HTML>"))))

(defvar jazz-html-faces-alist
  '((jazz-keyword-face     . ("<FONT COLOR=\"#0000ee\">" . "</FONT>"))
    (jazz-constant-face    . ("<FONT COLOR=\"#8080ff\">" . "</FONT>"))
    (jazz-declaration-face . ("<FONT COLOR=\"#cd0000\">" . "</FONT>"))
    (jazz-link-face        . ("<FONT COLOR=\"#000000\"><U>" . "</U></FONT>"))
    (jazz-link-declaration-face
     . ("<FONT COLOR=\"#cd0000\"><U>" . "</U></FONT>"))
    (jazz-comment-face     . ("<FONT COLOR=\"#008b00\">" . "</FONT>"))
    (jazz-type-face        . ("<FONT COLOR=\"#b7860b\">" . "</FONT>"))
    (jazz-constraint-face  . ("<FONT COLOR=\"#e9967a\">" . "</FONT>"))
    (jazz-string-face      . ("<FONT COLOR=\"#a020f0\">" . "</FONT>"))))

(defun jazz-html-print-segment (close open beg end)
  (let (char)
    (if close (princ close))
    (if open  (princ open))
    (while (< beg end)
      (setq char (char-after beg))
      (cond
       ((= char ?<) (princ "&lt;"))
       ((= char ?>) (princ "&gt;"))
       ((= char ?&) (princ "&amp;"))
       ((= char ?\") (princ "&quot;"))
       ((= char ?\t) (princ "  "))
       (t (write-char char)))
      (setq beg (1+ beg)))))

(defun jazz-latex-listing ()
  "Formats current buffer in LaTeX"
  (interactive)
  (when (buffer-file-name) (jazz-fontify-buffer))
  (with-output-to-temp-buffer "*jazz-latex-listing*"
    (princ "\\begingroup")
    (princ "\\def\\color#1#2#3{\\special{ps: #1 #2 #3 setrgbcolor}}")
    (princ "\\def\\{{\\char`\\{}")
    (princ "\\def\\}{\\char`\\}}")
    (princ "\\def\\\\{\\char`\\\\}")
    (princ "\\begin{alltt}\n\n")
    (let ((beg 1)
          (end 0)
          face spec)
      (while (< end (point-max))
        (setq end (or (next-property-change beg) (point-max)))
        
        (let* ((face (get-text-property beg 'face))
               (color (or (and face
                               (cdr (assoc face jazz-latex-color-alist)))
                          "#000000")))
          (jazz-latex-print-segment color beg end)
          (setq beg end))))
    (princ "\\end{alltt}")
    (princ "\\endgroup")))

(defvar jazz-latex-color-alist
  '((jazz-keyword-face     . "#0000ee")
    (jazz-constant-face    . "#8080ff")
    (jazz-declaration-face . "#cd0000")
    (jazz-link-face        . "#000000")
    (jazz-link-declaration-face . "#cd0000")
    (jazz-comment-face     . "#008b00")
    (jazz-type-face        . "#b7860b")
    (jazz-constraint-face  . "#e9967a")
    (jazz-string-face      . "#a020f0")))

(defun jazz-color (color i)
  (/ (string-to-int (substring color (+ 1 (* 2 i)) (+ 3 (* 2 i))) 16) 255.0))
  
(defun jazz-latex-print-segment (color beg end)
  (let ((char))
    (princ "\\color{")
    (princ (jazz-color color 0))
    (princ "}{")
    (princ (jazz-color color 1))
    (princ "}{")
    (princ (jazz-color color 2))
    (princ "}")
    (while (< beg end)
      (setq char (char-after beg))
      (cond
       ((= char ?{) (princ "\\{"))
       ((= char ?}) (princ "\\}"))
       ((= char ?\\) (princ "\\\\"))
;       ((= char ?\n)
;        (princ "\n")
;        (save-excursion
;          (goto-char (1+ beg))
;          (let ((count (current-indentation)))
;            (while (> count 0)
;              (princ " ")
;              (setq count (1- count))))
;          (if (jazz-search-forward "[^ 	]" (point-max))
;              (setq beg (- (point) 2)))))
       (t (write-char char)))
      (setq beg (1+ beg)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          Internal functions/variables
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; dir must end with /
(defun jazz-start-process (dir cmd)
  (message "%s//%s" dir cmd)
  (let (process
        (default-directory dir))
    (setq process (apply #'start-process "jazz" "*Jazz*" (car cmd) (cdr cmd)))
    (set-process-filter process 'jazz-process-filter)
    (set-process-sentinel process 'jazz-process-sentinel)
    process))
 
(defvar jazz-compiling nil)
(or (assq 'jazz-compiling minor-mode-alist)
    (setq minor-mode-alist (cons '(jazz-compiling " Compiling")
                                 minor-mode-alist)))

(defun jazz-search-forward (regexp limit)
  (let ((case-fold-search nil))
    (re-search-forward regexp limit t)))

(defun jazz-search-backward (regexp limit)
  (let ((case-fold-search nil))
    (re-search-backward regexp limit t)))

;;; get rid of ^M on NT
(defun clean-insert(s)
  (if (eq window-system 'w32)
      (let ((max))
        (save-excursion
          (insert s)
          (setq max (point)))
        (while (jazz-search-forward (char-to-string 13) max)
          (goto-char (match-beginning 0))
          (delete-char 1)
          (setq max (- max 1)))
        (goto-char max))
    (insert s)))

(defun jazz-process-filter (process string)
  (save-excursion
    (set-buffer (process-buffer process))
    (goto-char (point-max))
    (save-excursion
      (clean-insert string))
    (beginning-of-line)
    (while (jazz-search-forward "%[0-9\"].*:.*\n" (point-max))
      (save-excursion
        (goto-char (match-beginning 0))
        (indent-to (+ (- (frame-width) (match-end 0)) (match-beginning 0)))))))

(defun jazz-process-sentinel (process change)
  (cond
   ((string-equal change "finished\n")
    (setq jazz-compiling nil))
   ((string-match "exited abnormally with code 2" change)
    (setq jazz-compiling nil)
    (beep)
    (message "Error."))
   ((string-match "exited abnormally with code" change)
    (setq jazz-compiling nil)
    (beep)
    (message "Compiler bug." change)))
  (save-excursion
    (set-buffer (process-buffer process))
    (toggle-read-only 1)))

(defun jazz-match-to-int (n)
  (string-to-int (match-string-no-properties n)))

(defun jazz-line-char-to-pos (buffer l c)
  (save-excursion
    (set-buffer buffer)
    (goto-line l)
    (beginning-of-line)
    (forward-char (1- c))
    (point)))

(defun jazz-point-on-current-line (point)
  (save-excursion
    (let ((bol (progn (beginning-of-line) (point))))
      (goto-char point)
      (beginning-of-line)
      (= bol (point)))))
  
(defun jazz-select-file (name other l1 c1 l2 c2)
  (if (string-match "^[0-9]+$" name)
      (save-excursion
        (set-buffer "*Jazz*")
        (goto-char (point-max))
        (if (jazz-search-backward
             (concat name
                     ": %file:[ ]*\"\\([^\"]+\\)\" \"\\([^\"]+\\)\"")
             (point-min))
            (setq name (concat (match-string-no-properties 1) "/"
                               (match-string-no-properties 2))))))
  (if (not (file-name-absolute-p name))
      (setq name (concat jazz-directory name)))
  (let ((buffer (and (file-exists-p name)
                     (find-file-noselect name))))
    (if (not buffer)
        (message "Cannot find file \"%s\"" name (beep))
      (setq c1 (jazz-line-char-to-pos buffer l1 c1))
      (setq c2 (1+ (jazz-line-char-to-pos buffer l2 c2)))
      (if other
          (pop-to-buffer buffer)
        (switch-to-buffer buffer))
      (if (not (get-buffer-window buffer))
          (message "Cannot select a buffer for \"%s\"" name)
        (jazz-select (get-buffer-window buffer) c1 c2)))))

(defun jazz-no-more-errors ()
  (beep)
  (message "No more errors."))

(defun jazz-locate ()
  "Display the location indicated by the pragma at the end of the line."
  (interactive)
  (let ((window (selected-window))
        bol loc file l1 c1 l2 c2)
    (save-excursion
      (beginning-of-line)
      (setq bol (point))
      (end-of-line)
      (cond
       ((jazz-search-backward jazz-file-location-regexp bol)
        (jazz-select-file
         (match-string-no-properties 1)
         t
         (jazz-match-to-int 2)
         (jazz-match-to-int 3)
         (jazz-match-to-int 4)
         (jazz-match-to-int 5))
        (select-window window))
       ((jazz-search-backward "%\\([0-9]+\\)%" bol)
        (save-excursion
          (setq loc (concat "[^0-9]"
                            (match-string-no-properties 1)
                            ": %loc:[ ]*"
                            jazz-loc-regexp))
          (goto-char (point-max))
          (if (jazz-search-backward loc (point-min))
              (let ((file (match-string-no-properties 1))
                    (l1 (jazz-match-to-int 2))
                    (c1 (jazz-match-to-int 3))
                    (l2 (jazz-match-to-int 4))
                    (c2 (jazz-match-to-int 5)))
                (jazz-select-file file t l1 c1 l2 c2)
                (select-window window)))))))))

(provide 'jazz-mode)
